# Copyright (C) 2017 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.




import json
import os

from lib.base import (
    FLOWSPEC_NAME_TO_TYPE,
    BGPContainer,
    CmdBuffer,
    try_several_times,
    wait_for_completion,
    yellow,
    indent,
    local,
)


class YABGPContainer(BGPContainer):

    WAIT_FOR_BOOT = 1
    SHARED_VOLUME = '/etc/yabgp'

    def __init__(self, name, asn, router_id,
                 ctn_image_name='osrg/yabgp:v0.4.0'):
        super(YABGPContainer, self).__init__(name, asn, router_id,
                                             ctn_image_name)
        self.shared_volumes.append((self.config_dir, self.SHARED_VOLUME))

    def _copy_helper_app(self):
        import lib
        mod_dir = os.path.dirname(lib.__file__)
        local('docker cp {0}/yabgp_helper.py'
              ' {1}:/root/'.format(mod_dir, self.docker_name()))

    def _start_yabgp(self):
        self.local(
            'python /root/yabgp_helper.py'
            ' --config-file {0}/yabgp.ini'.format(self.SHARED_VOLUME),
            detach=True)

    def _wait_for_boot(self):
        return wait_for_completion(self._curl_is_running)

    def run(self):
        super(YABGPContainer, self).run()
        # self.create_config() is called in super class
        self._copy_helper_app()
        # To start YABGP, it is required to configure neighbor settings, so
        # here does not start YABGP yet.
        # self._start_yabgp()
        # self._wait_for_boot()
        return self.WAIT_FOR_BOOT

    def create_config(self):
        # Currently, supports only single peer
        c = CmdBuffer('\n')
        c << '[DEFAULT]'
        c << 'log_dir = {0}'.format(self.SHARED_VOLUME)
        c << 'use_stderr = False'
        c << '[message]'
        c << 'write_disk = True'
        c << 'write_dir = {0}/data/bgp/'.format(self.SHARED_VOLUME)
        c << 'format = json'

        if self.peers:
            info = next(iter(list(self.peers.values())))
            remote_as = info['remote_as']
            neigh_addr = info['neigh_addr'].split('/')[0]
            local_as = info['local_as'] or self.asn
            local_addr = info['local_addr'].split('/')[0]
            c << '[bgp]'
            c << 'afi_safi = ipv4, ipv6, vpnv4, vpnv6, flowspec, evpn'
            c << 'remote_as = {0}'.format(remote_as)
            c << 'remote_addr = {0}'.format(neigh_addr)
            c << 'local_as = {0}'.format(local_as)
            c << 'local_addr = {0}'.format(local_addr)

        with open('{0}/yabgp.ini'.format(self.config_dir), 'w') as f:
            print(yellow('[{0}\'s new yabgp.ini]'.format(self.name)))
            print(yellow(indent(str(c))))
            f.writelines(str(c))

    def reload_config(self):
        if self.peers == 0:
            return

        def _reload():
            def _is_running():
                ps = self.local('ps -ef', capture=True)
                running = False
                for line in ps.split('\n'):
                    if 'yabgp_helper' in line:
                        running = True
                return running

            if _is_running():
                self.local('/usr/bin/pkill -9 python')

            self._start_yabgp()
            self._wait_for_boot()
            if not _is_running():
                raise RuntimeError()

        try_several_times(_reload)

    def _curl_is_running(self):
        c = CmdBuffer(' ')
        c << "curl -X GET"
        c << "-u admin:admin"
        c << "-H 'Content-Type: application/json'"
        c << "http://localhost:8801/v1/"
        c << "> /dev/null 2>&1; echo $?"
        return self.local(str(c), capture=True) == '0'

    def _curl_send_update(self, path, peer):
        c = CmdBuffer(' ')
        c << "curl -X POST"
        c << "-u admin:admin"
        c << "-H 'Content-Type: application/json'"
        c << "http://localhost:8801/v1/peer/{0}/send/update".format(peer)
        c << "-d '{0}'".format(json.dumps(path))
        return json.loads(self.local(str(c), capture=True))

    def _construct_ip_unicast_update(self, rf, prefix, nexthop):
        # YABGP v0.4.0
        #
        # IPv4 Unicast:
        # curl -X POST \
        #      -u admin:admin \
        #      -H 'Content-Type: application/json' \
        #      http://localhost:8801/v1/peer/172.17.0.2/send/update -d '{
        #     "attr": {
        #         "1": 0,
        #         "2": [
        #             [
        #                 2,
        #                 [
        #                     1,
        #                     2,
        #                     3
        #                 ]
        #             ]
        #         ],
        #         "3": "192.0.2.1",
        #         "5": 500
        #     },
        #     "nlri": [
        #         "172.20.1.0/24",
        #         "172.20.2.0/24"
        #     ]
        # }'
        #
        # IPv6 Unicast:
        # curl -X POST \
        #      -u admin:admin \
        #      -H 'Content-Type: application/json' \
        #      http://localhost:8801/v1/peer/172.17.0.2/send/update -d '{
        #     "attr": {
        #         "1": 0,
        #         "2": [
        #             [
        #                 2,
        #                 [
        #                     65502
        #                 ]
        #             ]
        #         ],
        #         "4": 0,
        #         "14": {
        #             "afi_safi": [
        #                 2,
        #                 1
        #             ],
        #             "linklocal_nexthop": "fe80::c002:bff:fe7e:0",
        #             "nexthop": "2001:db8::2",
        #             "nlri": [
        #                 "::2001:db8:2:2/64",
        #                 "::2001:db8:2:1/64",
        #                 "::2001:db8:2:0/64"
        #             ]
        #         }
        #     }
        # }'
        if rf == 'ipv4':
            return {
                "attr": {
                    "3": nexthop,
                },
                "nlri": [prefix],
            }
        elif rf == 'ipv6':
            return {
                "attr": {
                    "14": {  # MP_REACH_NLRI
                        "afi_safi": [2, 1],
                        "nexthop": nexthop,
                        "nlri": [prefix],
                    },
                },
            }
        else:
            raise ValueError(
                'invalid address family for ipv4/ipv6 unicast: %s' % rf)

    def _construct_ip_unicast_withdraw(self, rf, prefix):
        # YABGP v0.4.0
        #
        # IPv4 Unicast:
        # curl -X POST \
        #      -u admin:admin \
        #      -H 'Content-Type: application/json' \
        #      http://localhost:8801/v1/peer/172.17.0.2/send/update -d '{
        #     "withdraw": [
        #         "172.20.1.0/24",
        #         "172.20.2.0/24"
        #     ]
        # }'
        #
        # IPv6 Unicast:
        # curl -X POST \
        #      -u admin:admin \
        #      -H 'Content-Type: application/json' \
        #      http://localhost:8801/v1/peer/172.17.0.2/send/update -d '{
        #     "attr": {
        #         "15": {
        #             "afi_safi": [
        #                 2,
        #                 1
        #             ],
        #             "withdraw": [
        #                 "::2001:db8:2:2/64",
        #                 "::2001:db8:2:1/64",
        #                 "::2001:db8:2:0/64"
        #             ]
        #         }
        #     }
        # }'
        if rf == 'ipv4':
            return {
                "withdraw": [prefix],
            }
        elif rf == 'ipv6':
            return {
                "attr": {
                    "15": {  # MP_UNREACH_NLRI
                        "afi_safi": [2, 1],
                        "withdraw": [prefix],
                    },
                },
            }
        else:
            raise ValueError(
                'invalid address family for ipv4/ipv6 unicast: %s' % rf)

    def _construct_flowspec_match(self, matchs):
        assert isinstance(matchs, (tuple, list))
        ret = {}
        for m in matchs:
            # m = "source-port '!=2 !=22&!=222'"
            # typ = "source-port"
            # args = "'!=2 !=22&!=222'"
            typ, args = m.split(' ', 1)
            # t = 6
            t = FLOWSPEC_NAME_TO_TYPE.get(typ, None)
            if t is None:
                raise ValueError('invalid flowspec match type: %s' % typ)
            # args = "!=2|!=22&!=222"
            args = args.strip("'").strip('"').replace(' ', '|')
            ret[t] = args
        return ret

    def _construct_flowspec_update(self, rf, matchs, thens):
        # YABGP v0.4.0
        #
        # curl -X POST \
        #      -u admin:admin \
        #      -H 'Content-Type: application/json' \
        #      http://localhost:8801/v1/peer/172.17.0.2/send/update -d '{
        #     "attr": {
        #         "1": 0,
        #         "14": {
        #             "afi_safi": [
        #                 1,
        #                 133
        #             ],
        #             "nexthop": "",
        #             "nlri": [
        #                 {
        #                     "1": "10.0.0.0/24"
        #                 }
        #             ]
        #         },
        #         "16": [
        #             "traffic-rate:0:0"
        #         ],
        #         "2": [],
        #         "5": 100
        #     }
        # }'
        #
        # Format of "thens":
        #   "traffic-rate:<AS>:<rate>"
        #   "traffic-marking-dscp:<int value>"
        #   "redirect-nexthop:<int value>"
        #   "redirect-vrf:<RT>"
        thens = thens or []
        if rf == 'ipv4-flowspec':
            afi_safi = [1, 133]
        else:
            raise ValueError('invalid address family for flowspec: %s' % rf)

        return {
            "attr": {
                "14": {  # MP_REACH_NLRI
                    "afi_safi": afi_safi,
                    "nexthop": "",
                    "nlri": [self._construct_flowspec_match(matchs)]
                },
                "16": thens,  # EXTENDED COMMUNITIES
            },
        }

    def _construct_flowspec_withdraw(self, rf, matchs):
        # curl -X POST \
        #      -u admin:admin \
        #      -H 'Content-Type: application/json' \
        #      http://localhost:8801/v1/peer/172.17.0.2/send/update -d '{
        #     "attr": {
        #         "15": {
        #             "afi_safi": [
        #                 1,
        #                 133
        #             ],
        #             "withdraw": [
        #                 {
        #                     "1": "192.88.2.3/24",
        #                     "2": "192.89.1.3/24"
        #                 },
        #                 {
        #                     "1": "192.88.4.3/24",
        #                     "2": "192.89.2.3/24"
        #                 }
        #             ]
        #         }
        #     }
        # }'
        if rf == 'ipv4-flowspec':
            afi_safi = [1, 133]
        else:
            raise ValueError('invalid address family for flowspec: %s' % rf)

        return {
            "attr": {
                "15": {  # MP_UNREACH_NLRI
                    "afi_safi": afi_safi,
                    "withdraw": [self._construct_flowspec_match(matchs)],
                },
            },
        }

    def _update_path_attributes(self, path, aspath=None, med=None,
                                local_pref=None):
        # ORIGIN: Currently support only IGP(0)
        path['attr']['1'] = 0
        # AS_PATH: Currently support only AS_SEQUENCE(2)
        if aspath is None:
            path['attr']['2'] = []
        else:
            path['attr']['2'] = [[2, aspath]]
        # MED
        if med is not None:
            path['attr']['4'] = med
        # LOCAL_PREF
        if local_pref is not None:
            path['attr']['5'] = local_pref
        # TODO:
        # Support COMMUNITY and EXTENDED COMMUNITIES

        return path

    def add_route(self, route, rf='ipv4', attribute=None, aspath=None,
                  community=None, med=None, extendedcommunity=None,
                  nexthop=None, matchs=None, thens=None,
                  local_pref=None, identifier=None, reload_config=True):
        self.routes.setdefault(route, [])

        for info in list(self.peers.values()):
            peer = info['neigh_addr'].split('/')[0]

            if rf in ['ipv4', 'ipv6']:
                nexthop = nexthop or info['local_addr'].split('/')[0]
                path = self._construct_ip_unicast_update(
                    rf, route, nexthop)
            # TODO:
            # Support "evpn" address family
            elif rf in ['ipv4-flowspec', 'ipv6-flowspec']:
                path = self._construct_flowspec_update(
                    rf, matchs, thens)
            else:
                raise ValueError('unsupported address family: %s' % rf)

            self._update_path_attributes(
                path, aspath=aspath, med=med, local_pref=local_pref)

            self._curl_send_update(path, peer)

        self.routes[route].append({
            'prefix': route,
            'rf': rf,
            'attr': attribute,
            'next-hop': nexthop,
            'as-path': aspath,
            'community': community,
            'med': med,
            'local-pref': local_pref,
            'extended-community': extendedcommunity,
            'identifier': identifier,
            'matchs': matchs,
            'thens': thens,
        })

    def del_route(self, route, identifier=None, reload_config=True):
        new_paths = []
        withdraw = None
        for p in self.routes.get(route, []):
            if p['identifier'] != identifier:
                new_paths.append(p)
            else:
                withdraw = p

        if not withdraw:
            return
        rf = withdraw['rf']

        for info in list(self.peers.values()):
            peer = info['neigh_addr'].split('/')[0]

            if rf in ['ipv4', 'ipv6']:
                r = self._construct_ip_unicast_withdraw(rf, route)
            elif rf == 'ipv4-flowspec':
                # NOTE: "ipv6-flowspec" does not seem to be supported with
                # YABGP v0.4.0
                matchs = withdraw['matchs']
                r = self._construct_flowspec_withdraw(rf, matchs)
            else:
                raise ValueError('unsupported address family: %s' % rf)

            self._curl_send_update(r, peer)

        self.routes[route] = new_paths

    def _get_adj_rib(self, peer, in_out='in'):
        peer_addr = self.peer_name(peer)
        c = CmdBuffer(' ')
        c << "curl -X GET"
        c << "-u admin:admin"
        c << "-H 'Content-Type: application/json'"
        c << "http://localhost:8801/v1-ext/peer/{0}/adj-rib-{1}".format(
            peer_addr, in_out)
        return json.loads(self.local(str(c), capture=True))

    def get_adj_rib_in(self, peer, rf='ipv4'):
        # "rf" should be either of;
        #   ipv4, ipv6, vpnv4, vpnv6, flowspec, evpn
        # The same as supported "afi_safi" in yabgp.ini
        ribs = self._get_adj_rib(peer, 'in')
        return ribs.get(rf, {})

    def get_adj_rib_out(self, peer, rf='ipv4'):
        # "rf" should be either of;
        #   ipv4, ipv6, vpnv4, vpnv6, flowspec, evpn
        # The same as supported "afi_safi" in yabgp.ini
        ribs = self._get_adj_rib(peer, 'out')
        return ribs.get(rf, {})



import time


def _get_link_local_address(ctn, ifname):
    out = ctn.local(
        'ip -6 -o addr show dev {0} scope link'.format(ifname),
        capture=True,
    )
    for line in out.split('\n'):
        if ' fe80:' not in line:
            continue
        return line.split()[3].split('/')[0]
    raise Exception('link local address not found')

def _ensure_neighbor_reachable(ctn, ifname, target_lladdr):
    out = ctn.local('ip -6 n show dev {0}'.format(ifname), capture=True)
    lines = [line for line in out.split('\n') if line.startswith('fe80')]
    for line in lines:
        lladdr = line.split()[0]
        if lladdr == target_lladdr:
            return 'REACHABLE' in line
    return False


# probe_link_local_address discovers the IPv6 link local address of the
# interfaces for container a and b connected p2p (or bridge network that only a
# and b exists).
def probe_link_local_address(a, b, aif, bif):
    done = False

    a_lladdr = _get_link_local_address(a, aif)
    b_lladdr = _get_link_local_address(b, bif)

    for i in range(20):
        a.local('ping6 -c 1 {0}%{1}'.format(b_lladdr, aif))
        b.local('ping6 -c 1 {0}%{1}'.format(a_lladdr, bif))
        if _ensure_neighbor_reachable(a, aif, b_lladdr) and _ensure_neighbor_reachable(b, bif, a_lladdr):
            done = True
            break
        time.sleep(1)

    if not done:
        raise Exception('timeout')



#!/usr/bin/env python3


import json
import logging
import sys

from flask import Blueprint
import flask

from yabgp.agent import prepare_service
from yabgp.common import constants
from yabgp.handler import BaseHandler
from yabgp.api import app
from yabgp.api import utils as api_utils
from yabgp.api.v1 import auth

LOG = logging.getLogger(__name__)
blueprint = Blueprint('v1-ext', __name__)
_send_update = api_utils.send_update


def _extract_afi_safi(attr):
    # MP_REACH_NLRI(14) or MP_UNREACH_NLRI(15)
    nlri = attr.get(14, None) or attr.get(15, None)
    if nlri and 'afi_safi' in nlri:
        afi_safi = nlri.get('afi_safi', None)
        return constants.AFI_SAFI_DICT.get(tuple(afi_safi), None)
    return 'ipv4'


def _construct_msg(attr, nlri, withdraw):
    return {
        'afi_safi': _extract_afi_safi(attr),
        'attr': attr,
        'nlir': nlri or [],
        'withdraw': withdraw or [],
    }


def _extract_nlri_list(msg):
    try:
        if msg['afi_safi'] == 'ipv4':
            return msg['nlri']
        return msg['attr'][14]['nlri']  # MP_REACH_NLRI(14)
    except KeyError:
        # Ignore case when no nlri
        pass
    return []


def _extract_withdraw_list(msg):
    try:
        if msg['afi_safi'] == 'ipv4':
            return msg['withdraw']
        return msg['attr'][15]['withdraw']  # MP_UNREACH_NLRI(15)
    except KeyError:
        # Ignore case when no withdraw
        pass
    return []


def _nlri_key(nlri):
    if isinstance(nlri, (dict, list)):
        return json.dumps(nlri)
    return str(nlri)


# ADJ_RIB_IN = {
#     <peer.factory.peer_addr>: {
#         <afi_safi>: {
#             <nlri>: <msg>,
#             ...
#         },
#         ...
#     },
#     ...
# }
ADJ_RIB_IN = {}


@blueprint.route('/peer/<peer_ip>/adj-rib-in')
@auth.login_required
@api_utils.log_request
def peer_adj_rib_in(peer_ip):
    """
    Dumps one peer's adj-RIB-in.
    """
    return flask.jsonify(ADJ_RIB_IN.get(peer_ip, {}))


# ADJ_RIB_OUT = {
#     <peer_ip>: {
#         <afi_safi>: {
#             <nlri>: <msg>,
#             ...
#         },
#         ...
#     },
#     ...
# }
ADJ_RIB_OUT = {}


@blueprint.route('/peer/<peer_ip>/adj-rib-out')
@auth.login_required
@api_utils.log_request
def peer_adj_rib_out(peer_ip):
    """
    Dumps one peer's adj-RIB-out.
    """
    return flask.jsonify(ADJ_RIB_OUT.get(peer_ip, {}))


app.app.register_blueprint(blueprint, url_prefix='/v1-ext')


def send_update(peer_ip, attr, nlri, withdraw):
    """
    Wrapper of "yabgp.api.send_update" in order to hook sending UPDATE
    messages via REST API.
    """
    msg = _construct_msg(attr, nlri, withdraw)
    afi_safi = msg['afi_safi']
    ADJ_RIB_OUT.setdefault(peer_ip, {})
    ADJ_RIB_OUT[peer_ip].setdefault(afi_safi, {})
    rib = ADJ_RIB_OUT[peer_ip][afi_safi]
    for _nlri in _extract_nlri_list(msg):
        rib[_nlri_key(_nlri)] = msg
    for _withdraw in _extract_withdraw_list(msg):
        rib.pop(_nlri_key(_withdraw), None)
    return _send_update(peer_ip, attr, nlri, withdraw)


setattr(api_utils, 'send_update', send_update)


class CliHandler(BaseHandler):

    def __init__(self):
        super(CliHandler, self).__init__()

    def init(self):
        pass

    def on_update_error(self, peer, timestamp, msg):
        LOG.info('[-] UPDATE ERROR: %s', msg)

    def route_refresh_received(self, peer, msg, msg_type):
        LOG.info('[+] ROUTE_REFRESH received: %s', msg)

    def keepalive_received(self, peer, timestamp):
        LOG.debug('[+] KEEPALIVE received: %s', peer.factory.peer_addr)

    def open_received(self, peer, timestamp, result):
        LOG.info('[+] OPEN received: %s', result)

    def update_received(self, peer, timestamp, msg):
        LOG.info('[+] UPDATE received: %s', msg)
        peer_addr = peer.factory.peer_addr
        afi_safi = msg['afi_safi']
        ADJ_RIB_IN.setdefault(peer.factory.peer_addr, {})
        ADJ_RIB_IN[peer_addr].setdefault(afi_safi, {})
        rib = ADJ_RIB_IN[peer_addr][afi_safi]
        for nlri in _extract_nlri_list(msg):
            rib[_nlri_key(nlri)] = msg
        for withdraw in _extract_withdraw_list(msg):
            rib.pop(_nlri_key(withdraw), None)

    def notification_received(self, peer, msg):
        LOG.info('[-] NOTIFICATION received: %s', msg)

    def on_connection_lost(self, peer):
        LOG.info('[-] CONNECTION lost: %s', peer.factory.peer_addr)

    def on_connection_failed(self, peer, msg):
        LOG.info('[-] CONNECTION failed: %s', msg)


def main():
    try:
        cli_handler = CliHandler()
        prepare_service(handler=cli_handler)
    except Exception as e:
        print(e)


if __name__ == '__main__':
    sys.exit(main())



# Copyright (C) 2016 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import time
import unittest

import collections
collections.Callable = collections.abc.Callable


from lib.noseplugin import parser_option

from lib import base
from lib.base import (
    BGP_FSM_ACTIVE,
    BGP_FSM_ESTABLISHED,
    wait_for_completion,
    local,
)
from lib.gobgp import GoBGPContainer


class GoBGPTestBase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        gobgp_ctn_image_name = parser_option.gobgp_image
        base.TEST_PREFIX = parser_option.test_prefix

        g1 = GoBGPContainer(name='g1', asn=65001, router_id='192.168.0.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level,
                            config_format='yaml')
        g2 = GoBGPContainer(name='g2', asn=65001, router_id='192.168.0.2',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level,
                            config_format='yaml')
        g3 = GoBGPContainer(name='g3', asn=65001, router_id='192.168.0.3',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level,
                            config_format='yaml')
        g4 = GoBGPContainer(name='g4', asn=65001, router_id='192.168.0.4',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level,
                            config_format='yaml')
        g5 = GoBGPContainer(name='g5', asn=65001, router_id='192.168.0.5',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level,
                            config_format='yaml')

        ctns = [g1, g2, g3, g4, g5]

        initial_wait_time = max(ctn.run() for ctn in ctns)

        time.sleep(initial_wait_time)

        g3.local("gobgp vrf add red rd 10:10 rt both 10:10")
        g3.local("gobgp vrf add blue rd 20:20 rt both 20:20")

        g1.add_peer(g3, graceful_restart=True, llgr=True)
        g3.add_peer(g1, vrf='red', is_rr_client=True, graceful_restart=True, llgr=True)

        g2.add_peer(g3, graceful_restart=True, llgr=True)
        g3.add_peer(g2, vrf='red', is_rr_client=True, graceful_restart=True, llgr=True)

        g4.add_peer(g3, graceful_restart=True, llgr=True)
        g3.add_peer(g4, vrf='blue', is_rr_client=True, graceful_restart=True, llgr=True)

        g5.add_peer(g3, graceful_restart=True, llgr=True)
        g3.add_peer(g5, vrf='blue', is_rr_client=True, graceful_restart=True, llgr=True)

        cls.g1 = g1
        cls.g2 = g2
        cls.g3 = g3
        cls.g4 = g4
        cls.g5 = g5

    # test each neighbor state is turned establish
    def test_01_neighbor_established(self):
        self.g3.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g1)
        self.g3.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g2)
        self.g3.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g4)
        self.g3.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g5)

    def test_02_add_routes(self):
        self.g1.local("gobgp global rib add 10.0.0.0/24")
        self.g4.local("gobgp global rib add 10.0.0.0/24")
        wait_for_completion(lambda: len(self.g3.get_global_rib(rf="vpnv4")) == 2)
        wait_for_completion(lambda: len(self.g2.get_global_rib()) == 1)
        wait_for_completion(lambda: len(self.g5.get_global_rib()) == 1)

    def test_03_disable(self):
        self.g3.disable_peer(self.g1)
        wait_for_completion(lambda: len(self.g3.get_global_rib(rf="vpnv4")) == 1)
        wait_for_completion(lambda: len(self.g2.get_global_rib()) == 0)
        wait_for_completion(lambda: len(self.g5.get_global_rib()) == 1)
        self.g3.enable_peer(self.g1)
        self.g3.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g1)

    def test_04_softreset_in(self):
        self.g3.softreset(self.g1)
        wait_for_completion(lambda: len(self.g3.get_global_rib()) == 0)
        wait_for_completion(lambda: len(self.g3.get_global_rib(rf="vpnv4")) == 2)
        wait_for_completion(lambda: len(self.g2.get_global_rib()) == 1)
        wait_for_completion(lambda: len(self.g5.get_global_rib()) == 1)

    def test_05_softreset_out(self):
        self.g3.softreset(self.g2, type='out')
        wait_for_completion(lambda: len(self.g3.get_global_rib()) == 0)
        wait_for_completion(lambda: len(self.g3.get_global_rib(rf="vpnv4")) == 2)
        wait_for_completion(lambda: len(self.g2.get_global_rib()) == 1)
        wait_for_completion(lambda: len(self.g5.get_global_rib()) == 1)

    def test_06_graceful_restart(self):
        self.g1.stop_gobgp()
        self.g3.wait_for(expected_state=BGP_FSM_ACTIVE, peer=self.g1)

        wait_for_completion(lambda: len(self.g3.get_global_rib(rf="vpnv4")) == 2)
        wait_for_completion(lambda: len(self.g2.get_global_rib()) == 1)

        wait_for_completion(lambda: len(self.g3.get_global_rib(rf="vpnv4")) == 1)
        wait_for_completion(lambda: len(self.g2.get_global_rib()) == 0)





# Copyright (C) 2015 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import time
import unittest

import collections
collections.Callable = collections.abc.Callable


from lib.noseplugin import parser_option

from lib import base
from lib.base import (
    BGP_FSM_IDLE,
    BGP_FSM_ESTABLISHED,
    BGP_ATTR_TYPE_COMMUNITIES,
    local,
)
from lib.gobgp import GoBGPContainer
from lib.exabgp import ExaBGPContainer


def community_exists(path, com):
    a, b = com.split(':')
    com = (int(a) << 16) + int(b)
    for a in path['attrs']:
        if a['type'] == BGP_ATTR_TYPE_COMMUNITIES and com in a['communities']:
            return True
    return False


class GoBGPTestBase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        gobgp_ctn_image_name = parser_option.gobgp_image
        base.TEST_PREFIX = parser_option.test_prefix

        g1 = GoBGPContainer(name='g1', asn=65000, router_id='192.168.0.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level,
                            config_format=parser_option.config_format)
        q1 = ExaBGPContainer(name='q1', asn=65001, router_id='192.168.0.2')
        q2 = ExaBGPContainer(name='q2', asn=65002, router_id='192.168.0.3')
        q3 = ExaBGPContainer(name='q3', asn=65003, router_id='192.168.0.4')

        qs = [q1, q2, q3]
        ctns = [g1, q1, q2, q3]

        initial_wait_time = max(ctn.run() for ctn in ctns)
        time.sleep(initial_wait_time)

        g1.local('gobgp global policy export add default reject')

        for q in qs:
            g1.add_peer(q)
            q.add_peer(g1)

        # advertise a route from q1, q2, q3
        for idx, q in enumerate(qs):
            route = '10.0.{0}.0/24'.format(idx + 1)
            q.add_route(route)

        cls.gobgp = g1
        cls.quaggas = {'q1': q1, 'q2': q2, 'q3': q3}

    # test each neighbor state is turned establish
    def test_01_neighbor_established(self):
        for q in self.quaggas.values():
            self.gobgp.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=q)

    def test_02_check_adj_rib_out(self):
        for q in self.quaggas.values():
            self.assertEqual(len(self.gobgp.get_adj_rib_out(q)), 0)

    def test_03_add_peer(self):
        q = ExaBGPContainer(name='q4', asn=65004, router_id='192.168.0.5')
        time.sleep(q.run())
        self.gobgp.add_peer(q)
        q.add_peer(self.gobgp)
        q.add_route('10.10.0.0/24')
        self.gobgp.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=q)
        self.quaggas['q4'] = q
        for q in self.quaggas.values():
            self.assertEqual(len(self.gobgp.get_adj_rib_out(q)), 0)

    def test_04_disable_peer(self):
        q3 = self.quaggas['q3']
        self.gobgp.disable_peer(q3)
        self.gobgp.wait_for(expected_state=BGP_FSM_IDLE, peer=q3)

        for q in self.quaggas.values():
            if q.name == 'q3':
                continue
            self.assertEqual(len(self.gobgp.get_adj_rib_out(q)), 0)

    def test_05_enable_peer(self):
        q3 = self.quaggas['q3']
        self.gobgp.enable_peer(q3)
        self.gobgp.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=q3)

        for q in self.quaggas.values():
            self.assertEqual(len(self.gobgp.get_adj_rib_out(q)), 0)

    def test_06_disable_peer2(self):
        q3 = self.quaggas['q3']
        # advertise a route which was also advertised by q1
        # this route will be best for g1 because q3's router-id is larger
        # than q1
        q3.add_route('10.0.1.0/24')
        time.sleep(3)
        # then disable q3
        self.gobgp.disable_peer(q3)
        self.gobgp.wait_for(expected_state=BGP_FSM_IDLE, peer=q3)

        for q in self.quaggas.values():
            if q.name == 'q3':
                continue
            self.assertEqual(len(self.gobgp.get_adj_rib_out(q)), 0)

    def test_07_adv_to_one_peer(self):
        self.gobgp.local('gobgp policy neighbor add ns0 {0}'.format(self.gobgp.peers[self.quaggas['q1']]['neigh_addr'].split('/')[0]))
        self.gobgp.local('gobgp policy statement add st0')
        self.gobgp.local('gobgp policy statement st0 add condition neighbor ns0')
        self.gobgp.local('gobgp policy statement st0 add action accept')
        self.gobgp.local('gobgp policy add p0 st0')
        self.gobgp.local('gobgp global policy export add p0 default reject')
        for q in self.quaggas.values():
            self.gobgp.softreset(q, type='out')

    def test_08_check_adj_rib_out(self):
        for q in self.quaggas.values():
            if q.name == 'q3':
                continue
            paths = self.gobgp.get_adj_rib_out(q)
            if q == self.quaggas['q1']:
                self.assertEqual(len(paths), 2)
            else:
                self.assertEqual(len(paths), 0)

    def test_09_change_global_policy(self):
        self.gobgp.local('gobgp policy statement st0 add action community add 65100:10')
        self.gobgp.local('gobgp global policy export set p0 default accept')
        for q in self.quaggas.values():
            self.gobgp.softreset(q, type='out')

    def test_10_check_adj_rib_out(self):
        for q in self.quaggas.values():
            if q.name == 'q3':
                continue
            paths = self.gobgp.get_adj_rib_out(q)
            if q != self.quaggas['q3']:
                self.assertEqual(len(paths), 2)
            for path in paths:
                if q == self.quaggas['q1']:
                    self.assertTrue(community_exists(path, '65100:10'))
                else:
                    self.assertFalse(community_exists(path, '65100:10'))

    def test_11_add_ibgp_peer(self):
        q = ExaBGPContainer(name='q5', asn=65000, router_id='192.168.0.6')
        time.sleep(q.run())
        self.quaggas['q5'] = q

        self.gobgp.add_peer(q)
        q.add_peer(self.gobgp)
        self.gobgp.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=q)

    def test_12_add_local_pref_policy(self):
        self.gobgp.local('gobgp policy statement st1 add action accept')
        self.gobgp.local('gobgp policy statement st1 add action local-pref 300')
        self.gobgp.local('gobgp policy add p1 st1')
        self.gobgp.local('gobgp global policy export set p1 default reject')
        for q in self.quaggas.values():
            self.gobgp.softreset(q, type='out')

    def test_13_check_adj_rib_out(self):
        q1 = self.quaggas['q1']
        for path in self.gobgp.get_adj_rib_out(q1):
            self.assertTrue(path['local-pref'] is None)
        q5 = self.quaggas['q5']
        for path in self.gobgp.get_adj_rib_out(q5):
            self.assertEqual(path['local-pref'], 300)

    def test_14_route_type_condition_local(self):
        self.gobgp.local('gobgp policy statement st2 add action accept')
        self.gobgp.local('gobgp policy statement st2 add condition route-type local')
        self.gobgp.local('gobgp policy add p2 st2')
        self.gobgp.local('gobgp global policy export set p2 default reject')
        for q in self.quaggas.values():
            self.gobgp.softreset(q, type='out')

        q1 = self.quaggas['q1']
        self.assertEqual(len(self.gobgp.get_adj_rib_out(q1)), 0)

        self.gobgp.add_route('10.20.0.0/24')

        time.sleep(1)

        self.assertEqual(len(self.gobgp.get_adj_rib_out(q1)), 1)
        self.assertEqual(self.gobgp.get_adj_rib_out(q1)[0]['nlri']['prefix'], '10.20.0.0/24')

    def test_15_route_type_condition_internal(self):
        self.gobgp.local('gobgp policy statement st22 add action accept')
        self.gobgp.local('gobgp policy statement st22 add condition route-type internal')
        self.gobgp.local('gobgp policy add p2 st22')
        self.gobgp.local('gobgp policy del p2 st2')
        for q in self.quaggas.values():
            self.gobgp.softreset(q, type='out')

        q1 = self.quaggas['q1']
        self.assertEqual(len(self.gobgp.get_adj_rib_out(q1)), 0)

        q5 = self.quaggas['q5']
        q5.add_route('10.30.0.0/24')

        time.sleep(1)

        self.assertEqual(len(self.gobgp.get_adj_rib_out(q1)), 1)
        self.assertEqual(self.gobgp.get_adj_rib_out(q1)[0]['nlri']['prefix'], '10.30.0.0/24')

    def test_16_route_type_condition_external(self):
        self.gobgp.local('gobgp policy statement st222 add action accept')
        self.gobgp.local('gobgp policy statement st222 add condition route-type external')
        self.gobgp.local('gobgp policy add p2 st222')
        self.gobgp.local('gobgp policy del p2 st22')
        for q in self.quaggas.values():
            self.gobgp.softreset(q, type='out')

        q1 = self.quaggas['q1']
        num1 = len(self.gobgp.get_adj_rib_out(q1))

        self.gobgp.add_route('10.40.0.0/24')
        time.sleep(1)
        num2 = len(self.gobgp.get_adj_rib_out(q1))
        self.assertEqual(num1, num2)

        q5 = self.quaggas['q5']
        q5.add_route('10.50.0.0/24')
        time.sleep(1)
        num3 = len(self.gobgp.get_adj_rib_out(q1))
        self.assertEqual(num1, num3)

        q2 = self.quaggas['q2']
        q2.add_route('10.60.0.0/24')
        time.sleep(1)
        num4 = len(self.gobgp.get_adj_rib_out(q1))
        self.assertEqual(num1 + 1, num4)

    def test_17_multi_statement(self):
        self.gobgp.local('gobgp policy statement st3 add action med set 100')
        self.gobgp.local('gobgp policy statement st4 add action local-pref 100')
        self.gobgp.local('gobgp policy add p3 st3 st4')
        self.gobgp.local('gobgp global policy import set p3 default accept')

        self.gobgp.add_route('10.70.0.0/24')
        time.sleep(1)
        rib = self.gobgp.get_global_rib('10.70.0.0/24')
        self.assertEqual(len(rib), 1)
        self.assertEqual(len(rib[0]['paths']), 1)
        path = rib[0]['paths'][0]
        self.assertEqual(path['med'], 100)
        self.assertEqual(path['local-pref'], 100)

    def test_18_reject_policy(self):
        self.gobgp.local('gobgp global policy import set default reject')
        self.gobgp.local('gobgp neighbor all softresetin')

        time.sleep(1)

        # self-generated routes remain since softresetin doesn't re-evaluate
        # them
        for v in self.gobgp.get_global_rib():
            for p in v['paths']:
                self.assertEqual(p['nexthop'], '0.0.0.0')





# Copyright (C) 2015 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import time
import unittest

import collections
collections.Callable = collections.abc.Callable


from lib.noseplugin import parser_option

from lib import base
from lib.base import BGP_FSM_ESTABLISHED, local
from lib.gobgp import GoBGPContainer
from lib.quagga import QuaggaBGPContainer


def wait_for(f, timeout=120):
    interval = 1
    count = 0
    while True:
        if f():
            return

        time.sleep(interval)
        count += interval
        if count >= timeout:
            raise Exception('timeout')


class GoBGPTestBase(unittest.TestCase):
    def assert_adv_count(self, src, dst, rf, count):
        self.assertEqual(count, len(src.get_adj_rib_out(dst, rf=rf)))
        self.assertEqual(count, len(dst.get_adj_rib_in(src, rf=rf)))

    def assert_upd_count(self, src, dst, sent, received):
        messages = src.get_neighbor(dst)['state']['messages']
        self.assertEqual(messages['sent'].get('update', 0), sent)
        self.assertEqual(messages['received'].get('update', 0), received)

    @classmethod
    def setUpClass(cls):
        gobgp_ctn_image_name = parser_option.gobgp_image
        base.TEST_PREFIX = parser_option.test_prefix

        g1 = GoBGPContainer(name='g1', asn=65000, router_id='192.168.0.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)
        q1 = QuaggaBGPContainer(name='q1', asn=65000, router_id='192.168.0.2')
        q2 = QuaggaBGPContainer(name='q2', asn=65000, router_id='192.168.0.3')
        q3 = QuaggaBGPContainer(name='q3', asn=65000, router_id='192.168.0.4')
        q4 = QuaggaBGPContainer(name='q4', asn=65000, router_id='192.168.0.5')

        qs = [q1, q2, q3, q4]
        ctns = [g1, q1, q2, q3, q4]

        initial_wait_time = max(ctn.run() for ctn in ctns)
        time.sleep(initial_wait_time)

        # g1 as a route reflector
        g1.add_peer(q1, is_rr_client=True)
        q1.add_peer(g1)
        g1.add_peer(q2, is_rr_client=True)
        q2.add_peer(g1)
        g1.add_peer(q3)
        q3.add_peer(g1)
        g1.add_peer(q4)
        q4.add_peer(g1)

        # advertise a route from q1, q2
        for idx, c in enumerate(qs):
            route = '10.0.{0}.0/24'.format(idx + 1)
            c.add_route(route)

        cls.gobgp = g1
        cls.quaggas = {'q1': q1, 'q2': q2, 'q3': q3, 'q4': q4}

    # test each neighbor state is turned establish
    def test_01_neighbor_established(self):
        for q in self.quaggas.values():
            self.gobgp.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=q)

    def test_02_check_gobgp_global_rib(self):
        for q in self.quaggas.values():
            # paths expected to exist in gobgp's global rib
            def f():
                state = self.gobgp.get_neighbor_state(q)
                self.assertEqual(state, BGP_FSM_ESTABLISHED)

                routes = list(q.routes.keys())
                global_rib = [p['prefix'] for p in self.gobgp.get_global_rib()]
                for p in global_rib:
                    if p in routes:
                        routes.remove(p)

                return len(routes) == 0
            wait_for(f)

    def test_03_check_gobgp_adj_rib_out(self):
        for q in self.quaggas.values():
            paths = [p['nlri']['prefix'] for p in self.gobgp.get_adj_rib_out(q)]
            for qq in self.quaggas.values():
                if q == qq:
                    continue
                if self.gobgp.peers[q]['is_rr_client']:
                    for p in list(qq.routes.keys()):
                        self.assertTrue(p in paths)
                else:
                    for p in list(qq.routes.keys()):
                        if self.gobgp.peers[qq]['is_rr_client']:
                            self.assertTrue(p in paths)
                        else:
                            self.assertFalse(p in paths)

    def test_10_setup_rr_rtc_isolation_policy(self):
        #                              +-------+
        #                              |  rr   |
        #        +----------------+----| (RR)  |---+----------------+
        #        |                |    +-------+   |                |
        #        |                |                |                |
        #      (iBGP)           (iBGP)           (iBGP)          (iBGP)
        #        |                |                |                |
        # +-------------+  +-------------+  +-------------+  +-------------+
        # |     acme1   |  |    acme2    |  |   tyrell1   |  |   tyrell2   |
        # | (RR Client) |  | (RR Client) |  | (RR Client) |  | (RR Client) |
        # +-------------+  +-------------+  +-------------+  +-------------+

        gobgp_ctn_image_name = parser_option.gobgp_image
        rr = GoBGPContainer(name='rr', asn=65000, router_id='192.168.1.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)
        acme1 = GoBGPContainer(name='acme1', asn=65000, router_id='192.168.1.101',
                               ctn_image_name=gobgp_ctn_image_name,
                               log_level=parser_option.gobgp_log_level)
        acme2 = GoBGPContainer(name='acme2', asn=65000, router_id='192.168.1.102',
                               ctn_image_name=gobgp_ctn_image_name,
                               log_level=parser_option.gobgp_log_level)

        tyrell1 = GoBGPContainer(name='tyrell1', asn=65000, router_id='192.168.1.201',
                                 ctn_image_name=gobgp_ctn_image_name,
                                 log_level=parser_option.gobgp_log_level)

        tyrell2 = GoBGPContainer(name='tyrell2', asn=65000, router_id='192.168.1.202',
                                 ctn_image_name=gobgp_ctn_image_name,
                                 log_level=parser_option.gobgp_log_level)

        time.sleep(max(ctn.run() for ctn in [rr, acme1, acme2, tyrell1, tyrell2]))

        rr.add_peer(
            acme1,
            vpn=True,
            addpath=16,
            graceful_restart=True,
            llgr=True,
            is_rr_client=True,
        )
        acme1.add_peer(rr, vpn=True, addpath=16, graceful_restart=True, llgr=True)

        rr.add_peer(
            acme2,
            vpn=True,
            addpath=16,
            graceful_restart=True,
            llgr=True,
            is_rr_client=True,
        )
        acme2.add_peer(rr, vpn=True, addpath=16, graceful_restart=True, llgr=True)

        rr.add_peer(
            tyrell1,
            vpn=True,
            addpath=16,
            graceful_restart=True,
            llgr=True,
            is_rr_client=True,
        )
        tyrell1.add_peer(rr, vpn=True, addpath=16, graceful_restart=True, llgr=True)

        rr.add_peer(
            tyrell2,
            vpn=True,
            addpath=16,
            graceful_restart=True,
            llgr=True,
            is_rr_client=True,
        )
        tyrell2.add_peer(rr, vpn=True, addpath=16, graceful_restart=True, llgr=True)

        self.__class__.rr = rr
        self.__class__.acme1 = acme1
        self.__class__.acme2 = acme2
        self.__class__.tyrell1 = tyrell1
        self.__class__.tyrell2 = tyrell2

        # add import/export policy to allow peers exchange routes within specific RTs
        # later tests should not break due to RTC Updates being filtered-out

        rr.local("gobgp policy neighbor add clients-acme {} {}".format(
            rr.peer_name(acme1),
            rr.peer_name(acme2)))

        rr.local("gobgp policy neighbor add clients-tyrell {} {}".format(
            rr.peer_name(tyrell1),
            rr.peer_name(tyrell2)))

        rr.local("gobgp policy ext-community add rts-acme   rt:^100:.*$")
        rr.local("gobgp policy ext-community add rts-tyrell rt:^200:.*$")

        rr.local('gobgp policy statement add allow-rtc')
        rr.local('gobgp policy statement allow-rtc add condition afi-safi-in rtc')
        rr.local('gobgp policy statement allow-rtc add action accept')

        rr.local('gobgp policy statement add allow-acme')
        rr.local('gobgp policy statement allow-acme add condition neighbor clients-acme')
        rr.local('gobgp policy statement allow-acme add condition ext-community rts-acme')
        rr.local('gobgp policy statement allow-acme add action accept')

        rr.local('gobgp policy statement add allow-tyrell')
        rr.local('gobgp policy statement allow-tyrell add condition neighbor clients-tyrell')
        rr.local('gobgp policy statement allow-tyrell add condition ext-community rts-tyrell')
        rr.local('gobgp policy statement allow-tyrell add action accept')
        rr.local('gobgp policy add tenancy allow-rtc allow-acme allow-tyrell')

        rr.local('gobgp global policy import add tenancy default reject')
        rr.local('gobgp global policy export add tenancy default reject')

        acme1.local("gobgp vrf add a1 rd 100:100 rt both 100:100")
        acme2.local("gobgp vrf add a1 rd 100:100 rt both 100:100")

        tyrell1.local("gobgp vrf add t1 rd 200:100 rt both 200:100")
        tyrell2.local("gobgp vrf add t1 rd 200:100 rt both 200:100")

        rr.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=acme1)
        rr.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=acme2)
        rr.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=tyrell1)
        rr.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=tyrell2)

    def test_11_routes_in_allowed_acme_rts_are_exchanged(self):
        self.acme1.local("gobgp vrf a1 rib add 10.10.0.0/16 local-pref 100")
        self.acme2.local("gobgp vrf a1 rib add 10.20.0.0/16")
        self.tyrell1.local("gobgp vrf t1 rib add 20.10.0.0/16")
        self.tyrell2.local("gobgp vrf t1 rib add 20.20.0.0/16")
        time.sleep(1)

        self.assert_adv_count(self.rr, self.acme1, 'rtc', 2)
        self.assert_adv_count(self.rr, self.acme1, 'ipv4-l3vpn', 1)
        self.assert_adv_count(self.rr, self.acme2, 'rtc', 2)
        self.assert_adv_count(self.rr, self.acme2, 'ipv4-l3vpn', 1)
        self.assert_adv_count(self.rr, self.tyrell1, 'rtc', 2)
        self.assert_adv_count(self.rr, self.tyrell1, 'ipv4-l3vpn', 1)
        self.assert_adv_count(self.rr, self.tyrell2, 'rtc', 2)
        self.assert_adv_count(self.rr, self.tyrell2, 'ipv4-l3vpn', 1)

    def test_12_routes_from_separate_rts_peers_are_isolated_by_rr(self):
        self.tyrell1.local("gobgp vrf add a1 rd 100:100 rt both 100:100")
        self.tyrell1.local("gobgp vrf a1 rib add 10.10.0.0/16 local-pref 200")
        self.tyrell1.local("gobgp vrf a1 rib add 10.30.0.0/16")
        time.sleep(1)

        rr_t2_in = self.rr.get_adj_rib_in(self.tyrell1, rf='ipv4-l3vpn')
        self.assertEqual(3, len(rr_t2_in))

        rr_a2_out = self.rr.get_adj_rib_out(self.acme2, rf='ipv4-l3vpn')
        self.assertEqual(1, len(rr_a2_out))

        a2_routes = self.acme2.get_adj_rib_in(self.rr, rf='ipv4-l3vpn')
        self.assertEqual(1, len(a2_routes))
        ar0 = a2_routes[0]
        self.assertEqual('10.10.0.0/16', ar0['prefix'])
        self.assertEqual(self.rr.peer_name(self.acme1), ar0['nexthop'])
        self.assertEqual(100, ar0['local-pref'])





# Copyright (C) 2016 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import time
import unittest

import collections
collections.Callable = collections.abc.Callable


from lib.noseplugin import parser_option

from lib import base
from lib.base import (
    BGP_FSM_IDLE,
    BGP_FSM_ACTIVE,
    BGP_FSM_ESTABLISHED,
    GRACEFUL_RESTART_TIME,
    local,
)
from lib.gobgp import GoBGPContainer


class GoBGPTestBase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        gobgp_ctn_image_name = parser_option.gobgp_image
        base.TEST_PREFIX = parser_option.test_prefix

        g1 = GoBGPContainer(name='g1', asn=65000, router_id='192.168.0.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)
        g2 = GoBGPContainer(name='g2', asn=65001, router_id='192.168.0.2',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)
        ctns = [g1, g2]

        initial_wait_time = max(ctn.run() for ctn in ctns)

        time.sleep(initial_wait_time)

        g1.add_route('10.10.10.0/24')
        g1.add_route('10.10.20.0/24')

        g1.add_peer(g2, graceful_restart=True)
        g2.add_peer(g1, graceful_restart=True)

        cls.bgpds = {'g1': g1, 'g2': g2}

    # test each neighbor state is turned establish
    def test_01_neighbor_established(self):
        g1 = self.bgpds['g1']
        g2 = self.bgpds['g2']
        g1.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=g2)

    def test_02_graceful_restart(self):
        g1 = self.bgpds['g1']
        g2 = self.bgpds['g2']
        g1.stop_gobgp()
        g2.wait_for(expected_state=BGP_FSM_ACTIVE, peer=g1)
        self.assertEqual(len(g2.get_global_rib('10.10.20.0/24')), 1)
        self.assertEqual(len(g2.get_global_rib('10.10.10.0/24')), 1)
        for d in g2.get_global_rib():
            for p in d['paths']:
                self.assertTrue(p['stale'])

        # Confirm the paths on the adj-RIB-in table are synced with the Global
        # table.
        self.assertEqual(len(g2.get_adj_rib_in(g1, '10.10.20.0/24')), 1)
        self.assertEqual(len(g2.get_adj_rib_in(g1, '10.10.10.0/24')), 1)
        for p in g2.get_adj_rib_in(g1):
            self.assertTrue(p['stale'])

        g1.routes = {}
        g1.start_gobgp(graceful_restart=True)
        g1.add_route('10.10.20.0/24')

    def test_03_neighbor_established(self):
        g1 = self.bgpds['g1']
        g2 = self.bgpds['g2']
        g1.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=g2)

        # Confirm the restart timer not expired.
        self.assertEqual(
            g2.local(
                "grep 'graceful restart timer expired' %s/gobgpd.log"
                " | wc -l" % (g2.SHARED_VOLUME), capture=True),
            '0')
        time.sleep(1)

        self.assertEqual(len(g2.get_global_rib('10.10.20.0/24')), 1)
        self.assertEqual(len(g2.get_global_rib('10.10.10.0/24')), 0)
        for d in g2.get_global_rib():
            for p in d['paths']:
                self.assertFalse(p.get('stale', False))

        # Confirm the stale paths are also removed from the adj-RIB-in table.
        # https://github.com/osrg/gobgp/pull/1707
        self.assertEqual(len(g2.get_adj_rib_in(g1, '10.10.20.0/24')), 1)
        self.assertEqual(len(g2.get_adj_rib_in(g1, '10.10.10.0/24')), 0)
        for p in g2.get_adj_rib_in(g1):
            self.assertFalse(p.get('stale', False))

    def test_04_add_non_graceful_restart_enabled_peer(self):
        g1 = self.bgpds['g1']
        # g2 = self.bgpds['g2']
        gobgp_ctn_image_name = parser_option.gobgp_image
        g3 = GoBGPContainer(name='g3', asn=65002, router_id='192.168.0.3',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)
        self.bgpds['g3'] = g3
        time.sleep(g3.run())
        g3.add_route('10.10.30.0/24')
        g1.add_peer(g3, graceful_restart=True)
        g3.add_peer(g1)
        g1.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=g3)
        time.sleep(1)
        self.assertEqual(len(g3.get_global_rib('10.10.20.0/24')), 1)
        self.assertEqual(len(g1.get_global_rib('10.10.30.0/24')), 1)

        # Restart g1 with graceful restart flag and check that g1 routes are still propagated to g3.
        g1.stop_gobgp()
        g3.stop_gobgp()
        g1.routes = {}
        g3.routes = {}
        g1.start_gobgp(graceful_restart=True)
        g1.add_route('10.10.20.0/24')
        g3.start_gobgp()
        # RFC 4724 Section 4.1: the Restarting Speaker MUST defer route
        # selection until it receives EOR from all peers that advertise GR
        # capability (excluding those with the Restart State bit set).
        # g2 advertises GR, so g1 must wait for g2's EOR before sending
        # routes to any peer, including g3 which does not advertise GR.
        g2 = self.bgpds['g2']
        g1.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=g2)
        g1.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=g3)
        time.sleep(1)
        self.assertEqual(len(g3.get_global_rib('10.10.20.0/24')), 1)
        g3.add_route('10.10.30.0/24')

    def test_05_holdtime_expiry_graceful_restart(self):
        g1 = self.bgpds['g1']
        g2 = self.bgpds['g2']
        g3 = self.bgpds['g3']
        g2.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=g1)
        g1.local("ip route add blackhole {}/32".format(g2.ip_addrs[0][1].split("/")[0]))
        g1.local("ip route add blackhole {}/32".format(g3.ip_addrs[0][1].split("/")[0]))
        g2.wait_for(expected_state=BGP_FSM_ACTIVE, peer=g1)
        self.assertEqual(len(g2.get_global_rib('10.10.20.0/24')), 1)
        self.assertEqual(len(g2.get_global_rib('10.10.30.0/24')), 1)
        for d in g2.get_global_rib():
            for p in d['paths']:
                self.assertTrue(p['stale'])

        self.assertEqual(len(g3.get_global_rib('10.10.20.0/24')), 0)
        self.assertEqual(len(g3.get_global_rib('10.10.30.0/24')), 1)

    def test_06_test_restart_timer_expire(self):
        g2 = self.bgpds['g2']
        time.sleep(GRACEFUL_RESTART_TIME + 5)
        g2 = self.bgpds['g2']
        self.assertEqual(len(g2.get_global_rib()), 0)

    def test_07_establish_after_graceful_restart(self):
        g1 = self.bgpds['g1']
        g2 = self.bgpds['g2']
        g3 = self.bgpds['g3']
        g1.local("ip route del blackhole {}/32".format(g2.ip_addrs[0][1].split("/")[0]))
        g1.local("ip route del blackhole {}/32".format(g3.ip_addrs[0][1].split("/")[0]))
        g1.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=g2)
        g1.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=g3)
        self.assertEqual(len(g2.get_global_rib('10.10.20.0/24')), 1)
        self.assertEqual(len(g2.get_global_rib('10.10.30.0/24')), 1)
        self.assertEqual(len(g3.get_global_rib('10.10.20.0/24')), 1)
        self.assertEqual(len(g3.get_global_rib('10.10.30.0/24')), 1)

    def test_08_graceful_restart(self):
        g1 = self.bgpds['g1']
        g2 = self.bgpds['g2']
        g3 = self.bgpds['g3']
        g1.stop_gobgp()
        g2.wait_for(expected_state=BGP_FSM_ACTIVE, peer=g1)
        self.assertEqual(len(g2.get_global_rib('10.10.20.0/24')), 1)
        self.assertEqual(len(g2.get_global_rib('10.10.30.0/24')), 1)
        for d in g2.get_global_rib():
            for p in d['paths']:
                self.assertTrue(p['stale'])

        self.assertEqual(len(g3.get_global_rib('10.10.20.0/24')), 0)
        self.assertEqual(len(g3.get_global_rib('10.10.30.0/24')), 1)

    def test_09_test_restart_timer_expire(self):
        time.sleep(GRACEFUL_RESTART_TIME + 5)
        g2 = self.bgpds['g2']
        self.assertEqual(len(g2.get_global_rib()), 0)

    def test_10_multineighbor_established(self):
        g1 = self.bgpds['g1']
        g2 = self.bgpds['g2']
        g3 = self.bgpds['g3']

        g1.start_gobgp()

        g1.del_peer(g2)
        g1.del_peer(g3)
        g2.del_peer(g1)
        g3.del_peer(g1)
        g1.add_peer(g2, graceful_restart=True, llgr=True)
        g1.add_peer(g3, graceful_restart=True, llgr=True)
        g2.add_peer(g1, graceful_restart=True, llgr=True)
        g3.add_peer(g1, graceful_restart=True, llgr=True)

        g2.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=g1)
        g3.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=g1)

    def test_11_multineighbor_graceful_restart(self):
        g1 = self.bgpds['g1']
        g2 = self.bgpds['g2']
        g3 = self.bgpds['g3']

        g1.stop_gobgp()
        g2.wait_for(expected_state=BGP_FSM_ACTIVE, peer=g1)
        g3.wait_for(expected_state=BGP_FSM_ACTIVE, peer=g1)

        g1.start_gobgp(graceful_restart=True)

        count = 0
        while (g1.get_neighbor_state(g2) != BGP_FSM_ESTABLISHED
               or g1.get_neighbor_state(g3) != BGP_FSM_ESTABLISHED):
            count += 1
            # assert connections are not refused
            self.assertTrue(g1.get_neighbor_state(g2) != BGP_FSM_IDLE)
            self.assertTrue(g1.get_neighbor_state(g3) != BGP_FSM_IDLE)
            if count > 120:
                raise Exception('timeout')
            time.sleep(1)





#!/usr/bin/env python

from __future__ import absolute_import
from __future__ import print_function

import grpc

import attribute_pb2
import common_pb2
import extcom_pb2
import gobgp_pb2
import gobgp_pb2_grpc
import nlri_pb2

_TIMEOUT_SECONDS = 1000


def go_bgp_subnet(color, endpoint_device, target_device, sid_list, bsid_value, nh):
    """
    inject or delete an route with <ACME>-CIDR and <ACME>-SCRUBBING community
    NLRI
    ORIGIN
    AS_PATH
    LP
    EXTENDED COMMUNITIES
     RT
    TUNNEL ENCAP
     TLVs
      SR Policy
       SUB-TLVs
        Preference
        Binding-SID
        SEG-LIST
         WEIGHT
         SEGMENT(1..n)
    """
    channel = grpc.insecure_channel("localhost:50051")
    stub = gobgp_pb2_grpc.GoBgpServiceStub(channel)

    family = common_pb2.Family(
        afi=common_pb2.Family.AFI_IP,
        safi=common_pb2.Family.SAFI_SR_POLICY,
    )

    nlri = nlri_pb2.NLRI(
        sr_policy=nlri_pb2.SRPolicyNLRI(
            color=color,
            distinguisher=2,
            endpoint=bytes(map(int, endpoint_device.split("."))),
            length=96,
        )
    )

    attributes = [
        attribute_pb2.Attribute(
            next_hop=attribute_pb2.NextHopAttribute(next_hop=nh)
        ),
        attribute_pb2.Attribute(
            origin=attribute_pb2.OriginAttribute(origin=0)
        ),
        attribute_pb2.Attribute(
            extended_communities=attribute_pb2.ExtendedCommunitiesAttribute(
                communities=[
                    extcom_pb2.ExtendedCommunity(
                        ipv4_address_specific=extcom_pb2.IPv4AddressSpecificExtended(
                            is_transitive=False,
                            sub_type=0x02,
                            address=target_device,
                            local_admin=0,
                        )
                    )
                ]
            )
        ),
    ]

    sr_binding_sid = attribute_pb2.SRBindingSID(
        s_flag=False,
        i_flag=False,
        sid=bsid_value.to_bytes(4, byteorder="big"),
    )

    segments = [
        attribute_pb2.TunnelEncapSubTLVSRSegmentList.Segment(
            a=attribute_pb2.SegmentTypeA(
                flags=attribute_pb2.SegmentFlags(s_flag=False),
                label=label << 12,
            )
        )
        for label in sid_list
    ]

    tunnel_tlv = attribute_pb2.TunnelEncapTLV(
        type=15,
        tlvs=[
            attribute_pb2.TunnelEncapTLV.TLV(
                sr_preference=attribute_pb2.TunnelEncapSubTLVSRPreference(
                    flags=0,
                    preference=11,
                )
            ),
            attribute_pb2.TunnelEncapTLV.TLV(
                sr_binding_sid=attribute_pb2.TunnelEncapSubTLVSRBindingSID(
                    sr_binding_sid=sr_binding_sid
                )
            ),
            attribute_pb2.TunnelEncapTLV.TLV(
                sr_segment_list=attribute_pb2.TunnelEncapSubTLVSRSegmentList(
                    weight=attribute_pb2.SRWeight(flags=0, weight=12),
                    segments=segments,
                )
            ),
        ],
    )

    attributes.append(
        attribute_pb2.Attribute(
            tunnel_encap=attribute_pb2.TunnelEncapAttribute(tlvs=[tunnel_tlv])
        )
    )

    stub.AddPath(
        gobgp_pb2.AddPathRequest(
            table_type=gobgp_pb2.TABLE_TYPE_GLOBAL,
            path=gobgp_pb2.Path(
                nlri=nlri,
                pattrs=attributes,
                family=family,
                best=True,
            ),
        ),
        timeout=_TIMEOUT_SECONDS,
    )


if __name__ == "__main__":
    nh = "10.100.1.201"  # gobgp ip
    endpoint_device = "10.6.6.6"  # https://datatracker.ietf.org/doc/html/draft-ietf-idr-segment-routing-te-policy-16#section-2.3
    color = 100
    target_device = "10.1.1.1"  # intended head-ends for the advertised SR Policy update
    bsid_value = 300004  # bsid
    sid_list = [200002, 200006]  # label stack
    go_bgp_subnet(
        color,
        endpoint_device=endpoint_device,
        target_device=target_device,
        bsid_value=bsid_value,
        sid_list=sid_list,
        nh=nh,
    )




# Copyright (C) 2016 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import time
import unittest

import collections
collections.Callable = collections.abc.Callable


from lib.noseplugin import parser_option

from lib import base
from lib.base import BGP_FSM_ESTABLISHED, local
from lib.gobgp import GoBGPContainer


class GoBGPTestBase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        gobgp_ctn_image_name = parser_option.gobgp_image
        base.TEST_PREFIX = parser_option.test_prefix

        g1 = GoBGPContainer(name='g1', asn=65001, router_id='192.168.0.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level,
                            config_format='yaml')
        g2 = GoBGPContainer(name='g2', asn=65002, router_id='192.168.0.2',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level,
                            config_format='yaml')
        g3 = GoBGPContainer(name='g3', asn=65003, router_id='192.168.0.3',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level,
                            config_format='yaml')
        g4 = GoBGPContainer(name='g4', asn=65004, router_id='192.168.0.4',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level,
                            config_format='yaml')
        g5 = GoBGPContainer(name='g5', asn=65005, router_id='192.168.0.5',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level,
                            config_format='yaml')
        g6 = GoBGPContainer(name='g6', asn=65006, router_id='192.168.0.6',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level,
                            config_format='yaml')
        g7 = GoBGPContainer(name='g7', asn=65007, router_id='192.168.0.7',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level,
                            config_format='yaml')

        ctns = [g1, g2, g3, g4, g5, g6, g7]

        initial_wait_time = max(ctn.run() for ctn in ctns)

        time.sleep(initial_wait_time)

        g4.local("gobgp vrf add red rd 10:10 rt both 10:10")
        g4.local("gobgp vrf add blue rd 20:20 rt both 20:20")

        g5.local("gobgp vrf add red rd 10:10 rt both 10:10")
        g5.local("gobgp vrf add blue rd 20:20 rt both 20:20")

        g1.add_peer(g4)
        g4.add_peer(g1, vrf='red')

        g2.add_peer(g4)
        g4.add_peer(g2, vrf='red')

        g3.add_peer(g4)
        g4.add_peer(g3, vrf='blue')

        g4.add_peer(g5, vpn=True)
        g5.add_peer(g4, vpn=True)

        g5.add_peer(g6, vrf='red')
        g6.add_peer(g5)

        g5.add_peer(g7, vrf='blue')
        g7.add_peer(g5)

        cls.g1 = g1
        cls.g2 = g2
        cls.g3 = g3
        cls.g4 = g4
        cls.g5 = g5
        cls.g6 = g6
        cls.g7 = g7

    # test each neighbor state is turned establish
    def test_01_neighbor_established(self):
        self.g4.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g1)
        self.g4.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g2)
        self.g4.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g3)
        self.g4.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g5)
        self.g5.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g6)
        self.g5.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g7)

    def test_02_inject_from_vrf_red(self):
        self.g1.local('gobgp global rib add 10.0.0.0/24')

        time.sleep(1)

        dst = self.g2.get_global_rib('10.0.0.0/24')
        self.assertEqual(len(dst), 1)
        self.assertEqual(len(dst[0]['paths']), 1)
        path = dst[0]['paths'][0]
        self.assertEqual([self.g4.asn, self.g1.asn], path['aspath'])

        dst = self.g3.get_global_rib('10.0.0.0/24')
        self.assertEqual(len(dst), 0)

        dst = self.g4.get_global_rib(rf='vpnv4')
        self.assertEqual(len(dst), 1)
        self.assertEqual(len(dst[0]['paths']), 1)
        path = dst[0]['paths'][0]
        self.assertEqual([self.g1.asn], path['aspath'])

        dst = self.g6.get_global_rib('10.0.0.0/24')
        self.assertEqual(len(dst), 1)
        self.assertEqual(len(dst[0]['paths']), 1)
        path = dst[0]['paths'][0]
        self.assertEqual([self.g5.asn, self.g4.asn, self.g1.asn], path['aspath'])

        dst = self.g7.get_global_rib('10.0.0.0/24')
        self.assertEqual(len(dst), 0)

    def test_03_inject_from_vrf_blue(self):
        self.g3.local('gobgp global rib add 10.0.0.0/24')

        time.sleep(1)

        dst = self.g2.get_global_rib('10.0.0.0/24')
        self.assertEqual(len(dst), 1)
        self.assertEqual(len(dst[0]['paths']), 1)
        path = dst[0]['paths'][0]
        self.assertEqual([self.g4.asn, self.g1.asn], path['aspath'])

        dst = self.g4.get_global_rib(rf='vpnv4')
        self.assertEqual(len(dst), 2)

        dst = self.g6.get_global_rib('10.0.0.0/24')
        self.assertEqual(len(dst), 1)
        self.assertEqual(len(dst[0]['paths']), 1)
        path = dst[0]['paths'][0]
        self.assertEqual([self.g5.asn, self.g4.asn, self.g1.asn], path['aspath'])

        dst = self.g7.get_global_rib('10.0.0.0/24')
        self.assertEqual(len(dst), 1)
        self.assertEqual(len(dst[0]['paths']), 1)
        path = dst[0]['paths'][0]
        self.assertEqual([self.g5.asn, self.g4.asn, self.g3.asn], path['aspath'])


